use std::collections::HashSet;

use futures::TryStreamExt;
use mongodb::{bson::doc, Collection};

// todo : api key auth

// case: itemset
// items describe generic top level structure

// case: item variants
// specific instance of an item (real world)
// possibly deviating attrs

// case: import
// made / bought an item -> supply
// check in with attrs -> uuid (transaction)
// commited to db

// case: export
// used an item -> demand
// mark as used
// update db

// ITEM
// VARIANTS
// QUANTIZATION
// IMPORT / EXPORT
// DEPENDENCIES
// DEMAND STATS
// SEASONAL REVIEWS

macro_rules! collect_results {
    ($res:expr) => {{
        let mut ret = vec![];

        while let Some(doc) = $res.try_next().await.unwrap() {
            ret.push(doc);
        }

        ret
    }};
}

macro_rules! get_mongo {
    () => {
        mongodb::Client::with_uri_str(std::env::var("DB_URI").unwrap())
            .await
            .unwrap()
    };
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemVariantEntry {
    pub item_name: String,
    pub variant_name: String,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemInventoryEntry {
    pub item_name: String,
    pub variant_name: String,
    pub origin: String,
    pub price: f64,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemEntry {
    pub name: String,
    pub category: String,
}

impl ItemEntry {
    pub fn new(doc: mdq::Document) -> Self {
        let name = std::path::Path::new(&doc.path)
            .file_stem()
            .unwrap()
            .to_str()
            .unwrap()
            .to_string();
        let category = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("category")
            .unwrap()
            .as_str()
            .unwrap()
            .to_string();
        Self { name, category }
    }
}

pub struct Item {
    db: mongodb::Client,
    pub item: ItemEntry,
}

impl Item {
    pub fn new(db: mongodb::Client, item: ItemEntry) -> Self {
        Self { db, item }
    }

    pub async fn get_inventory_entries(&self) -> Vec<ItemInventoryEntry> {
        let quantities: Collection<ItemInventoryEntry> =
            self.db.database("cdb").collection("item_quantities");

        let mut found = quantities
            .find(doc! { "item_name": self.item.name.clone()}, None)
            .await
            .unwrap();

        collect_results!(found)
    }

    pub async fn get_origins(&self) -> Vec<String> {
        let quantities: Collection<ItemInventoryEntry> =
            self.db.database("cdb").collection("item_quantities");

        let mut found = quantities
            .find(doc! { "item_name": self.item.name.clone()}, None)
            .await
            .unwrap();

        let mut hashset = HashSet::new();

        while let Some(res) = found.try_next().await.unwrap() {
            hashset.insert(res.origin);
        }

        hashset.into_iter().collect()
    }

    pub async fn add_inventory(&self, variant: &str, origin: &str, price: f64) -> String {
        // todo : implement
        let quantities: Collection<ItemInventoryEntry> =
            self.db.database("cdb").collection("item_quantities");
        let id = quantities
            .insert_one(
                ItemInventoryEntry {
                    item_name: self.item.name.clone(),
                    variant_name: variant.to_owned(),
                    origin: origin.to_owned(),
                    price,
                },
                None,
            )
            .await
            .unwrap();
        return id.inserted_id.as_object_id().unwrap().to_string();
    }

    pub async fn get_variants(&self) -> Vec<String> {
        let variants: Collection<ItemVariantEntry> =
            self.db.database("cdb").collection("item_variants");
        let mut found_variants = variants
            .find(
                doc! {
                    "item_name": self.item.name.clone()
                },
                None,
            )
            .await
            .unwrap();

        let mut ret = vec!["Generic".to_string()];

        while let Some(doc) = found_variants.try_next().await.unwrap() {
            ret.push(doc.variant_name);
        }

        ret
    }

    pub fn variant(&self, variant: &str) -> Option<Variant> {
        // todo : check if avail
        Some(Variant {
            item: self.item.name.clone(),
            variant: variant.to_string(),
        })
    }

    pub async fn add_variant(&self, name: &str) {
        let variants: Collection<ItemVariantEntry> =
            self.db.database("cdb").collection("item_variants");
        variants
            .insert_one(
                ItemVariantEntry {
                    item_name: self.item.name.clone(),
                    variant_name: name.to_owned(),
                },
                None,
            )
            .await
            .unwrap();
    }
}

#[derive(Debug, Clone)]
pub struct Price(String);

impl Price {
    fn new(price: &str) -> Self {
        // todo : implement

        Self(price.to_string())
    }
}

impl From<String> for Price {
    fn from(val: std::string::String) -> Self {
        Self::new(&val)
    }
}

pub struct Variant {
    pub item: String,
    pub variant: String,
}

pub struct Transaction {}

impl Transaction {
    pub fn new(item: &str, variant: &str, price: Price, origin: &str) -> Self {
        // todo : implement
        Self {}
    }

    fn as_doc(self) -> mongodb::bson::Document {
        mongodb::bson::doc! {}
    }
}

impl Variant {
    pub async fn supply(&self, amount: usize, price: Price, origin: &str) -> String {
        // todo : implement db transactions
        let db = get_mongo!();
        let col: mongodb::Collection<mongodb::bson::Document> =
            db.database("cdb").collection("transactions");

        let mut transactions = vec![];
        for _ in 0..amount {
            transactions.push(Transaction::new(
                &self.item,
                &self.variant,
                price.clone(),
                origin,
            ));
        }

        for transaction in transactions {
            let r = col.insert_one(transaction.as_doc(), None).await;
        }

        todo!()
    }
}

pub struct ItemDB {
    index: mdq::Index,
    mongodb: mongodb::Client,
}

impl ItemDB {
    pub async fn new(dir: &str, mongodb: &str) -> Self {
        // scan for markdown item entries
        let index = mdq::Index::new(dir, true);
        let mongodb = mongodb::Client::with_uri_str(mongodb).await.unwrap();
        /*
        for item in &index.documents {
            let item = ItemEntry::new(item.clone());
            log::info!("Adding item {} to DB", item.name);
            let items: Collection<ItemEntry> = mongodb.database("cdb").collection("items");
            if let None =
                items
                    .find_one(doc! { "name": item.name.clone()}, None)
                    .await
                    .unwrap()
            {
                items.insert_one(item, None).await.unwrap();
            } else {
                // todo : update values
            }
        }*/

        Self { index, mongodb }
    }

    /// Retrieves an item by name
    pub fn get_item(&self, item: &str) -> Option<Item> {
        Some(Item::new(
            self.mongodb.clone(),
            self.index
                .documents
                .iter()
                .map(|x| ItemEntry::new(x.clone()))
                .find(|x| x.name == item)?,
        ))
    }

    pub fn items(&self) -> Vec<String> {
        let mut ret = vec![];
        for item in &self.index.documents {
            let item = ItemEntry::new(item.clone());
            ret.push(item.name);
        }
        ret
    }
}
